--- old/usr/src/uts/common/fs/zfs/arc.c	Tue Aug  6 16:31:03 2013
+++ new/usr/src/uts/common/fs/zfs/arc.c	Tue Aug  6 16:31:03 2013
@@ -294,7 +294,6 @@
 	kstat_named_t arcstat_l2_writes_sent;
 	kstat_named_t arcstat_l2_writes_done;
 	kstat_named_t arcstat_l2_writes_error;
-	kstat_named_t arcstat_l2_writes_hdr_miss;
 	kstat_named_t arcstat_l2_evict_lock_retry;
 	kstat_named_t arcstat_l2_evict_reading;
 	kstat_named_t arcstat_l2_free_on_write;
@@ -360,7 +359,6 @@
 	{ "l2_writes_sent",		KSTAT_DATA_UINT64 },
 	{ "l2_writes_done",		KSTAT_DATA_UINT64 },
 	{ "l2_writes_error",		KSTAT_DATA_UINT64 },
-	{ "l2_writes_hdr_miss",		KSTAT_DATA_UINT64 },
 	{ "l2_evict_lock_retry",	KSTAT_DATA_UINT64 },
 	{ "l2_evict_reading",		KSTAT_DATA_UINT64 },
 	{ "l2_free_on_write",		KSTAT_DATA_UINT64 },
@@ -622,7 +620,7 @@
 uint64_t l2arc_feed_min_ms = L2ARC_FEED_MIN_MS;	/* min interval milliseconds */
 boolean_t l2arc_noprefetch = B_TRUE;		/* don't cache prefetch bufs */
 boolean_t l2arc_feed_again = B_TRUE;		/* turbo warmup */
-boolean_t l2arc_norw = B_TRUE;			/* no reads during writes */
+boolean_t l2arc_norw = B_FALSE;			/* no reads during writes */
 
 /*
  * L2ARC Internals
@@ -4148,7 +4146,7 @@
 	l2arc_dev_t *dev;
 	list_t *buflist;
 	arc_buf_hdr_t *head, *ab, *ab_prev;
-	l2arc_buf_hdr_t *abl2;
+	l2arc_buf_hdr_t *l2hdr;
 	kmutex_t *hash_lock;
 
 	cb = zio->io_private;
@@ -4170,26 +4168,19 @@
 	/*
 	 * All writes completed, or an error was hit.
 	 */
+	dev->l2ad_writing = B_FALSE;
 	for (ab = list_prev(buflist, head); ab; ab = ab_prev) {
 		ab_prev = list_prev(buflist, ab);
 
 		hash_lock = HDR_LOCK(ab);
-		if (!mutex_tryenter(hash_lock)) {
-			/*
-			 * This buffer misses out.  It may be in a stage
-			 * of eviction.  Its ARC_L2_WRITING flag will be
-			 * left set, denying reads to this buffer.
-			 */
-			ARCSTAT_BUMP(arcstat_l2_writes_hdr_miss);
-			continue;
-		}
+		mutex_enter(hash_lock);
 
-		abl2 = ab->b_l2hdr;
+		l2hdr = ab->b_l2hdr;
 
 		/*
 		 * Release the temporary compressed buffer as soon as possible.
 		 */
-		if (abl2->b_compress != ZIO_COMPRESS_OFF)
+		if (l2hdr->b_compress != ZIO_COMPRESS_OFF)
 			l2arc_release_cdata_buf(ab);
 
 		if (zio->io_error != 0) {
@@ -4197,9 +4188,9 @@
 			 * Error - drop L2ARC entry.
 			 */
 			list_remove(buflist, ab);
-			ARCSTAT_INCR(arcstat_l2_asize, -abl2->b_asize);
+			ARCSTAT_INCR(arcstat_l2_asize, -l2hdr->b_asize);
 			ab->b_l2hdr = NULL;
-			kmem_free(abl2, sizeof (l2arc_buf_hdr_t));
+			kmem_free(l2hdr, sizeof (l2arc_buf_hdr_t));
 			ARCSTAT_INCR(arcstat_l2_size, -ab->b_size);
 		}
 
@@ -4350,7 +4341,7 @@
 l2arc_evict(l2arc_dev_t *dev, uint64_t distance, boolean_t all)
 {
 	list_t *buflist;
-	l2arc_buf_hdr_t *abl2;
+	l2arc_buf_hdr_t *l2hdr;
 	arc_buf_hdr_t *ab, *ab_prev;
 	kmutex_t *hash_lock;
 	uint64_t taddr;
@@ -4450,10 +4441,10 @@
 			 * Tell ARC this no longer exists in L2ARC.
 			 */
 			if (ab->b_l2hdr != NULL) {
-				abl2 = ab->b_l2hdr;
-				ARCSTAT_INCR(arcstat_l2_asize, -abl2->b_asize);
+				l2hdr = ab->b_l2hdr;
+				ARCSTAT_INCR(arcstat_l2_asize, -l2hdr->b_asize);
 				ab->b_l2hdr = NULL;
-				kmem_free(abl2, sizeof (l2arc_buf_hdr_t));
+				kmem_free(l2hdr, sizeof (l2arc_buf_hdr_t));
 				ARCSTAT_INCR(arcstat_l2_size, -ab->b_size);
 			}
 			list_remove(buflist, ab);
@@ -4725,9 +4716,10 @@
 		dev->l2ad_first = B_FALSE;
 	}
 
+	/* dev->l2ad_writing will be lowered in the zio done callback */
 	dev->l2ad_writing = B_TRUE;
 	(void) zio_wait(pio);
-	dev->l2ad_writing = B_FALSE;
+	ASSERT(dev->l2ad_writing == B_FALSE);
 
 	return (write_asize);
 }
